from numpy import sqrt

def hex_to_rgb(hex_string: str) -> tuple[int, int, int]:
    hex_string = hex_string.lstrip("#")
    return tuple(int(hex_string[i:i+2], 16) for i in (0, 2, 4))


def lightness_coef(hex_color: str) -> float:
    """
    Luma handles lighter colors better, while weighted euclidean
    norm handles darker colors better. Start with weighted euclidean
    norm, and if the color is medium to light, switch to luma
    """
    coef = weighted_euclidean_norm_rgb(hex_to_rgb(hex_color))
    if coef > 0.4:
        return luma_rgb(hex_to_rgb(hex_color))
    return coef


def color_is_light(hex_color: str) -> bool:
    min_to_be_considered_light = 0.525
    coef = lightness_coef(hex_color)
    if coef > min_to_be_considered_light:
        return True
    return False


def get_alt_shade(color: str, shade_coef: float | None = 0.4) -> str:
    try:
        from bokeh.colors import RGB
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "To use auto-shading, `bokeh` must be installed. Please 'pip install bokeh'"
        )
    
    downscale_lightness_strength = 5
    strengthen_weight_exponential = 2  # power

    is_light = color_is_light(color)
    lightness = lightness_coef(color)
    lightness_strength = abs(lightness - 0.5)
    if is_light is False:
        strengthen_weight_exponential += ((lightness_strength + 1) ** 3) - 1
        # downscale_lightness_strength -= 1
    lightness_strength_scaled = lightness_strength / downscale_lightness_strength
    weight_exp = ((lightness_strength_scaled + 1) ** strengthen_weight_exponential) - 1

    weight = weight_exp
    # print(round(weight * 10, 3), "-", strengthen_weight_exponential, "-", round(lightness_strength * 10, 3), "-", round(lightness * 10, 3))

    rgb = RGB(*hex_to_rgb(color))

    if is_light:
        res = rgb.darken(shade_coef + weight)
    else:
        res = rgb.lighten(shade_coef + weight)
    
    return res.to_hex().lstrip("#")


def luma_rgb(rgb: list | tuple) -> float:
    """
    Luma of rgb vector / 255
    """
    r, g, b = rgb
    return (
        (
            (0.212 * r)
            + (0.701 * g)
            + (0.087 * b)
        )
        / 255
    )


def weighted_euclidean_norm_rgb(rgb: list | tuple) -> float:
    """
    Dear programmer: I have no idea how this works, but it determines
    how 'light' a color is.

    Returns: float between 0 and 1
    """
    r, g, b = rgb
    denom = 255 * sqrt(0.299 + 0.587 + 0.114)
    return sqrt(0.299 * r**2 + 0.587 * g**2 + 0.114 * b**2) / denom







import re
from excelbird._formulae import FORMULAE

def remove_paren_enclosure(value: str) -> str:
    if not isinstance(value, str):
        return value
    if value.startswith("(") and value.endswith(")"):
        return value.removeprefix("(").removesuffix(")")
    return value


def autofit_algorithm(value: str) -> int:
    """
    Decides column width given string value of a cell
    """
    filtered_value = str(value).replace("_xlfn.", "")
    length_coef = len(filtered_value)
    with_lower_bound = max(length_coef, 10)
    with_upper_bound = min(with_lower_bound, 40)
    return with_upper_bound


def approximate_arial_string_width(st):
    import string
    size = 0 # in milinches
    for s in st:
        if s in 'lij|\' ': size += 37
        elif s in '![]fI.,:;/\\t': size += 50
        elif s in '`-(){}r"': size += 60
        elif s in '*^zcsJkvxy': size += 85
        elif s in 'aebdhnopqug#$L+<>=?_~FZT' + string.digits: size += 95
        elif s in 'BSPEAKVXY&UwNRCHD': size += 112
        elif s in 'QGOMm%W@': size += 135
        else: size += 50
    return size * 6 / 1000.0 # Convert to picas


def format_formula(s: str) -> str:
    """
    The last step in finalizing a formula. Attempting to correct
    user mistakes, trailing commas, unwanted spaces, etc. Assume
    the function is fully built, starts with '=', and all functions
    are prefixed with '_xlfn.'
    """
    s = re.sub(r"\([, ]+", r"(", s)  # comma before opening paren
    s = re.sub(r"[, ]+\)", r")", s) # trailing comma before closing paren
    s = re.sub(r"^= +", "=", s)  # leading space
    s = re.sub(r"[, ]+$", "", s)  # trailing comma
    s = re.sub(r"= *'(.*?)'", r'="\1"', s)  # using single quotes instead of double
    return s


def prefix_formulae_funcs(func: str) -> str:
    """
    Openpyxl will insert '@' before any function it doesn't think is a valid
    excel function. Its list of functions is outdated, so newer functions, like "CONCAT"
    will be interpreted as invalid incorrectly.

    Our solution is to find functions in the string, where there are capital letters, dots,
    or digits, followed by a starting parenthese, check if they are in openpyxl's
    `FORMULAE`, and if not, prefix them with "_xlfn."
    """

    matches = re.findall(r"([A-Z\.a-z0-9]+)\(", func)
    # matches = re.findall(r"^([A-Z\.0-9]+)\(", func) + re.findall(r"(?<![^\.A-Za-z0-9])([A-Za-z\.0-9]+)\(", func)
    matches = set([match for match in matches if match.upper() in FORMULAE])
    for match in matches:
        func = func.replace(match + "(", "_xlfn." + match + "(")
    return func



"""
Detailed documentation and code examples coming soon.
"""
from __future__ import annotations
# External
from pandas import Series, DataFrame, concat
from numpy import ndarray
from typing import Any, Iterable
from copy import deepcopy
import re

from excelbird.styles.styles import default_table_style
from excelbird._base.container import ListIndexableById
from excelbird._base.identifier import HasId
from excelbird._base.styling import HasBorder
from excelbird.core.item import Item
from excelbird._base.dotdict import Style
from excelbird._base.loc import Loc
from excelbird._base.math import CanDoMath, elem_math

from excelbird._utils.util import (
    get_dimensions,
    init_from_same_dimension_type,
)
from excelbird._utils.argument_parsing import (
    combine_args_and_children_to_list,
    convert_all_to_type,
    move_remaining_kwargs_to_dict,
    convert_sibling_types,
)

from excelbird._utils.pass_attributes import (
    pass_dict_to_children,
)
from excelbird._utils.validation import (
    require_each_element_to_be_cls_type,
    ensure_value_is_not_number,
)

from excelbird.core.expression import Expr
from excelbird.core.function import Func

from excelbird.core.gap import Gap
from excelbird.core.series import (
    _Series,
    Col,
    Row,
)


class _Frame(CanDoMath, ListIndexableById, HasId, HasBorder):
    _doc_primary_summary = """
    A 2-dimensional vector that can :ref:`be used in a python expression <expression_main>`.
    """
    _doc_params = """
    Parameters
    ----------
    *args : Union[Col, Row, Frame, VFrame, list, tuple, pd.Series, pd.DataFrame, np.ndarray, Gap, Item, Expr, Func, set]
        Children must be (or resolve to) a series. Frame holds Cols, and VFrame
        holds Rows - `Gap` and `Item` will be interpreted as the respective element. Can also
        take any value that will be resolved to one of the above types, such as a list, tuple,
        pandas Series, etc. 2-dimensional arguments, such as pandas DataFrame, will be 'exploded'
        inplace into separate 1-dimensional elements.
    children : list, optional
        Will be combined with args
    id : str, optional
        Unique identifier to store globally so that this element can be referenced
        elsewhere in the layout without being assigned to a variable
    schema : Schema, optional
        A Schema object to use to rename child headers to desired output names at write time.
    sep : Gap or bool or int or dict, optional
        A sep in any excelbird layout element inserts a Gap between each of its children.
        If True, a default of ``Gap(1)`` is used. If int, ``Gap(sep)`` will be used. If a dict,
        ``Gap(1, **sep)`` will be used.
    sizes : dict[str, int], optional
        Specify the column width (or row height, if `VFrame`) for any child element by header.
        Keys should be the header of a child element, and values should be integers representing
        that element's size. Note: unlike most excelbird styling, this argument will override any
        other column widths / row heights given to the children.
    background_color : str, optional
        Hex code for background color. Will be applied to fill_color of any Gap child who hasn't specified its own
        fill_color. Will also be passed down to any Col/Row child who hasn't specified its own background_color.
    fill_empty : bool, optional
        Fill shorter children (if children vary in length) with ``Cell("")`` so that all lengths are matching,
        and all Cells inside the child will follow the same style. If False or None, these empty spaces will instead
        be filled with ``Gap()``, to which the child's background_color will be applied, if present.
    cell_style : dict, optional
        Will be applied to each child's cell_style
    header_style : dict, optional
        Will be applied to each child's header_style
    table_style : dict or bool, optional
        Format a Frame as an Excel table. (ignored for VFrame). If True, default style
        'name="TableStyleMedium2"' is used. If dict, key 'displayName' will be used as the
        table name, and all other key/values will be passed to openpyxl.worksheet.table.TableStyleInfo.
    border : list[tuple or str or bool] or tuple[str or bool, str or bool] or str or bool, optional
        Syntax inspired by CSS. A non-list value will be applied to all 4 sides. If list,
        length can be 2, 3, or 4 elements. Order is [top, right, bottom, left]. If length 2,
        apply the first element to top and bottom border, and apply the second element to right and left.
        To apply border to children instead, use cell_style.
    border_top : tuple[str or bool, str or bool] or str or bool, optional
        Top border. If True, a thin black border is used. If string (6 char hex code),
        use the default weight and apply the specified color. If string (valid weight name),
        use the default color and apply the specified weight. If tuple, apply the first
        element as weight, and second element as color.
    border_right : tuple[str or bool, str or bool] or str or bool, optional
        Right border. See border_top
    border_bottom : tuple[str or bool, str or bool] or str or bool, optional
        Bottom border. See border_top
    border_left : tuple[str or bool, str or bool] or str or bool, optional
        Left border. See border_top
    **kwargs : Any
        Remaining kwargs will be applied to cell_style

    """
    _dimensions = 2
    elem_type = _Series

    def __init__(
        self,
        *args: Any,
        children: list | None = None,
        id: str | int | None = None,
        schema: None = None,
        sep: Any | None = None,
        sizes: list | None = None,
        border_top: bool | str | None = None,
        border_right: bool | str | None = None,
        border_bottom: bool | str | None = None,
        border_left: bool | str | None = None,
        border: bool | str | Iterable | None = None,
        background_color: str | None = None,
        fill_empty: bool | None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        table_style: Style | dict | bool | None = None,
        **kwargs,
    ) -> None:
        children = combine_args_and_children_to_list(args, children)

        children = [i for i in children if i is not None]

        children = init_from_same_dimension_type(self, children)
        if getattr(self, "_id", None) is not None and id is None:
            id = self.id

        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()
        if table_style is None or table_style is False:
            table_style = dict()
        elif table_style is True:
            table_style = default_table_style

        self._format_args(children)

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        self._loc = None
        self.id = id
        self.schema = schema
        self.sizes = sizes
        self.background_color = background_color
        self.fill_empty = fill_empty
        self.header_style = Style(**header_style)
        self.table_style = Style(**table_style)
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)

        self._init(children)

        self._init_border(
            border,
            border_top,
            border_right,
            border_bottom,
            border_left,
        )
        if sep is not None:
            self._insert_separator(sep)

    @property
    def shape(self) -> tuple[int, int]:
        return (
            max([v.shape[0] for v in self if hasattr(v, "shape")] + [0]),
            sum([1 if not isinstance(i, Gap) else i for i in self]),
        )

    @property
    def headers(self) -> list[str | None]:
        """
        The headers of each child. `None` will be placed at the index
        of children who have no header, so the returned list will be of
        the same length as self.

        Returns
        -------
        list[str or None]
        """
        return [i.header if hasattr(i, "_header") else None for i in self]

    def ref(self, inherit_style: bool = False, **kwargs):
        """
        Get a new object with cell references to those in the caller.
        This assumes that **both** the calling object
        and the returned object will be placed in the workbook.

        .. note::

            Calling ``.ref()`` is **not** necessary when an object is used in
            a python expression (i.e. ``some_cell + some_row``) and should `only`
            be used to duplicate data across a workbook.

        Parameters
        ----------
        inherit_style : bool, default False
            Copy the caller's style to the returned object.

        Returns
        -------
        :class:`Frame <excelbird.Frame>` or :class:`VFrame <excelbird.VFrame>`
            Self type

        Notes
        -----

        .. note::

            Children's ``header`` attributes are stylistic attributes, and therefore will **not** be
            passed to the returned object's children unless ``inherit_style=True``. And, if style
            is inherited, headers will be copied over to the children, instead of cell references to them.

        """
        new_elements = [
            i.ref(inherit_style=inherit_style, **kwargs)
            if not isinstance(i, Gap)
            else deepcopy(i)
            for i in self
        ]
        new_dict = kwargs
        if inherit_style is True:
            self_dict = deepcopy(self.__dict__)
            for key, val in self_dict.items():
                if key == "_header":
                    key = "header"
                if key not in new_dict and key not in ["_id", "_loc"]:
                    new_dict[key] = val
        return type(self)(*new_elements, **new_dict)

    def transpose(self, **kwargs):
        """
        Convert to sibling type. Places current children into the returned object,
        without copying or making cell references to them.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to apply as attributes to the new object.

        Returns
        -------
        :class:`Frame <excelbird.Frame>` or :class:`VFrame <excelbird.VFrame>`
            The opposite to self's type. Try ``type(my_obj).sibling_type``

        Notes
        -----
        **Assumes that the caller won't be placed in the layout**. Do not
        place both the calling object and returned object in the layout, since
        they both contain the same children.

        .. code-block::

            # 'current' must not be placed in the workbook.
            new = current.transpose()

        To include both, use a reference of `current` instead

        .. code-block::

            new = current.ref().transpose()

        """
        elements = list(self)
        new = type(self).sibling_type(*elements)
        for key, val in self.__dict__.items():
            if key == "_id":
                key = "id"
            setattr(new, key, val)
        for key, val in kwargs.items():
            if hasattr(new, key):
                setattr(new, key, val)
            elif hasattr(new, 'cell_style'):
                new.cell_style[key] = val
        return new

    def range(self, include_headers: bool = False):
        """
        Get a reference to the entire range of the frame, instead of a vector of
        cell references.

        Parameters
        ----------
        include_headers : bool, default False
            If True, the header cells will be included in the range reference.

        Returns
        -------
        :class:`Cell <excelbird.Cell>`
        """

        if getattr(self[0], 'header_written', False) is True and include_headers is False:
            first = self[0][1]
        else:
            first = self[0][0]

        last = self[-1][-1]
        return first >> last

    def _format_args(self, args: list) -> None:
        self._explode_all_2d_iterables(args)
        convert_all_to_type(args, (Series, tuple, ndarray), type(self).elem_type)
        convert_all_to_type(args, list, type(self).elem_type, strict=True)
        convert_all_to_type(args, set, Expr)
        Item._resolve_all_in_container(args, type(self).elem_type)
        convert_sibling_types(self, args)
        for i, elem in enumerate(args):
            if not isinstance(elem, (type(self).elem_type, Gap, Func, Expr)):
                args[i] = type(self).elem_type(elem)

    def _explode_all_2d_iterables(self, args: list) -> None:
        for i, elem in enumerate(args):
            if isinstance(elem, DataFrame):
                df = args.pop(i)
                for col in reversed(df.columns):
                    args.insert(i, df[col])

            elif isinstance(elem, type(self)):
                frame = args.pop(i)
                for sr in reversed(frame):
                    args.insert(i, sr)

            elif isinstance(elem, ndarray):
                if len(elem.shape) == 2:
                    arr2d = args.pop(i)
                    for sr in reversed(arr2d):
                        args.insert(i, sr)

            elif type(elem) is list or isinstance(elem, tuple):
                if all(isinstance(e, (list, tuple, ndarray, Series, Item, Gap, Expr, Func)) for e in elem):
                    iterable2d = args.pop(i)
                    for sr in reversed(iterable2d):
                        args.insert(i, sr)

    def _resolve_background_color(self) -> None:
        for elem in self:
            if hasattr(elem, "_resolve_background_color"):
                if (
                    self.background_color not in [None, False]
                    and elem.background_color is None
                ):
                    elem.background_color = self.background_color
                elem._resolve_background_color()

        if self.background_color not in [None, False]:
            for elem in self:
                if isinstance(elem, Gap):
                    if "fill_color" not in elem.kwargs:
                        elem.fill = True
                        elem.kwargs["fill_color"] = self.background_color

    def _key_to_idx(self, key: str | int) -> int:
        try:
            return super()._key_to_idx(key)
        except (KeyError, IndexError):
            if key in self.headers:
                return self.headers.index(key)

            raise KeyError(f"Invalid key, {key}")

    def __getitem__(self, key):
        if isinstance(key, (int, str, slice)):
            return super().__getitem__(key)

        if not isinstance(key, list):
            # return super().__getitem__(key)
            return ListIndexableById.__getitem__(self, key)

        new_elements = [self[self._key_to_idx(k)] for k in key]
        new_dict = {k: v for k, v in self.__dict__.items() if k not in ["_id", "_loc"]}
        if "_header" in new_dict:
            new_dict["header"] = new_dict.pop("_header")

        return type(self)(*new_elements, **new_dict)

    def __setitem__(self, key, val) -> None:
        if isinstance(key, int):
            return super().__setitem__(key, val)
        if isinstance(val, Func):
            val.kwargs['header'] = key
        else:
            val.header = key
        try:
            index = self._key_to_idx(key)
            self[index] = val
        except Exception:
            self.append(val)

    def _set_loc(self, loc: Loc) -> None:
        self._loc = loc

        offset = self._starting_offset()
        for elem in self:
            elem._set_loc(
                Loc((self._loc.y + offset.y, self._loc.x + offset.x), self._loc.ws)
            )
            offset = self._inc_offset(offset, elem)

    def _apply_sizes(self) -> None:
        def set_elem_size(elem, size):
            if not isinstance(elem, (Col, Row)):
                return
            attr = "col_width" if isinstance(elem, Col) else "row_height"

            elem.cell_style[attr] = size
            if elem.header is not None:
                elem.header_style[attr] = size

        if isinstance(self.sizes, (list, tuple)):
            for elem, size in zip(self, self.sizes):
                set_elem_size(elem, size)

        elif isinstance(self.sizes, dict):
            for key, val in self.sizes.items():
                elem = self.get(key)
                if elem is not None:
                    set_elem_size(elem, val)

    def _validate_child_types(self) -> None:
        cls_name = type(self).__name__
        elem_type_name = type(self).elem_type.__name__
        valid_types = (
            type(self).elem_type,
            Gap,
        )
        for elem in self:
            if not isinstance(elem, valid_types):
                raise TypeError(
                    f"At write time, a {cls_name} can only hold {elem_type_name}s or Gaps. "
                    "To arrange mixed types, place them in a Stack or VStack"
                )
            if hasattr(elem, "_validate_child_types"):
                elem._validate_child_types()

    def _write(self) -> None:
        require_each_element_to_be_cls_type(self)
        self._apply_border()

        # Safely set each style to the element's header style, if it hasn't already
        # been set.
        pass_dict_to_children(self, "header_style")
        pass_dict_to_children(self, "cell_style")

        self._apply_sizes()

        for elem in self:
            # If a schema has been declared, check if the
            # value of the first cell is present in the schema. If it is,
            # replace the value with its output label in the schema
            if self.schema is not None:
                if len(elem) != 0:
                    schema_var = self.schema.get(elem.header)
                    if schema_var is not None:
                        elem.header = schema_var.output

            elem._write()

    def _resolve_gaps(self) -> None:
        Gap._explode_all_to_series(self, type(self).elem_type, self._gap_size)
        for elem in self:
            if hasattr(elem, "_resolve_gaps"):
                elem._resolve_gaps()

    # def __rshift__(self, other):
    #     if get_dimensions(other) < get_dimensions(self):
    #         return elem_math(self[0], other, lambda a, b: a >> b, " >> ")
    #     return self[0] >> other[-1]
    #
    # def __rrshift__(self, other):
    #     if get_dimensions(other) < get_dimensions(self):
    #         return elem_math(other, self[-1], lambda a, b: a >> b, " >> ")
    #     return other[0] >> self[-1]


class Frame(_Frame):

    _doc_custom_summary = """
    * Direction: **horizontal**
    * Child Type: :class:`Col`
    """

    sibling_type = None  # these are set after class declaration
    elem_type = Col

    def transpose(self, **kwargs) -> VFrame:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> Frame:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self) -> int:
        return self.shape[1]

    @property
    def height(self) -> int:
        return self.shape[0]

    def _write(self) -> None:
        if len(self.table_style) > 0:
            self._format_headers_for_table_format()

        super()._write()

        if len(self.table_style) > 0:
            self._apply_table_format_to_worksheet()

    def _format_headers_for_table_format(self) -> None:
        """
        To apply excel table format, all column headers must be strings,
        with no duplicates.

        1. Ensure all headers are strings. Throw error otherwise
        2. Add headers where missing:
            ('foo', None, 'bar', None) -> ('foo', 'Unnamed', 'bar', 'Unnamed2')
        3. Append counter to duplicate headers:
            ('a', 'b', 'b', 'b') -> ('a', 'b', 'b2', 'b3')

        Mutates inplace: `self`
        """
        for i, series in enumerate(self):
            if series.header is None:
                series.header = f"Unnamed"
            else:
                ensure_value_is_not_number(series.header)

        header_counts = dict()
        for i, series in enumerate(self):
            if series.header in header_counts:
                header_counts[series.header] += 1
            else:
                header_counts[series.header] = 1

        header_counts = {k: v for k, v in header_counts.items() if v > 1}

        for header in header_counts.keys():
            columns = [i for i in self if i.header == header]
            for i, col in enumerate(columns):
                if i > 0:
                    col.header += str(i + 1)

    def _apply_table_format_to_worksheet(self):
        """
        NOTE: THIS WILL FAIL IF HEADERS ARE REFERENCES FROM
        A DIFFERENT SHEET. To fix:
            Check for "!" in header names. If found, trace the
            cell_range expression back to the original cells and
            copy their values


        Formats self's cell range in worksheet as excel table

        Mutates inplace:
            `self._loc.ws`
            `self.table_style`
        """

        import openpyxl.worksheet.table as xl_tbl

        style = self.table_style
        cell_range = (
            self.range(include_headers=True)
            ._expr_value()
            .replace(self._loc.title_str, "")
        )
        ws = self._loc.ws

        if "displayName" in style:
            name = style.pop("displayName")
        else:
            if isinstance(ws.title, str):
                friendly_title = re.sub(r"[^A-Za-z]", "", ws.title)
                name = friendly_title + "Table1"
            else:
                name = "Table1"

        valid_table_name = True
        valid_table_name, attempts = False, 0
        err_msg = ""
        while valid_table_name is False and attempts < 30:
            attempts += 1
            try:
                table = xl_tbl.Table(displayName=name, ref=cell_range)
                table.tableStyleInfo = xl_tbl.TableStyleInfo(**style)
                ws.add_table(table)
                valid_table_name = True
            except Exception as e:
                err_msg = e
                name_and_num = re.search(r"(.+)(\d+)", name)
                if name_and_num is None:
                    name += "1"
                else:
                    label, num = name_and_num.groups()
                    name = f"{label}{int(num)+1}"

        if valid_table_name is False:
            raise ValueError(
                f"Couldn't properly format table on sheet, '{ws.title}', cell range '{cell_range}'."
                "This is either due to invalid data formatting, or a duplicate table name. Here is the "
                f"error message from openpyxl:\n{err_msg}"
            )

    @property
    def _gap_size(self) -> int:
        return self.height

    def _repr_html_(self):
        elements = [
            Series(
                list(e) if isinstance(e, list) else [e] + [""],
                name=e.header if getattr(e, "_header", None) is not None or getattr(e, 'header', None) is not None else "",
            )
            for e in self
        ]

        return (
            concat(elements, axis=1).fillna("").style.hide(axis="index")._repr_html_()
        )

    def _border_mask(self, top, right, bottom, left) -> Style:
        return Style(
            first=[top, False, bottom, left],
            last=[top, right, bottom, False],
            middle=[top, False, bottom, False],
        )

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.x += elem.width
        return offset

    def _starting_offset(self) -> Loc:
        offset = Loc((0, 0), self._loc.ws)
        if getattr(self, "_header", None) is not None:
            offset.x += 1
        return offset


class VFrame(_Frame):
    _doc_custom_summary = """
    * Direction: **vertical**
    * Child Type: :class:`Row`

    .. note:: Unlike :class:`Frame`, `VFrame` cannot be formatted as a table in Excel. Param ``table_style`` will be ignored.
    """

    sibling_type = None  # these are set after class declaration
    elem_type = Row

    def transpose(self, **kwargs) -> Frame:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> VFrame:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self) -> int:
        return self.shape[0]

    @property
    def height(self) -> int:
        return self.shape[1]

    @property
    def _gap_size(self) -> int:
        return self.width

    def _repr_html_(self):
        max_len = max([len(e) if isinstance(e, _Series) else 1 for e in self] + [0])
        elements = [
            Series(
                list(e) if isinstance(e, list) else [e],
                name=e.header if getattr(e, "_header", None) is not None or getattr(e, 'header', None) is not None else "",
            )
            for e in self
        ] + [Series(["" for _ in range(max_len)], name="")]

        df = DataFrame(elements)
        df.columns = ["" for _ in range(max([len(e) for e in elements] + [0]))]

        if not any(getattr(e, "_header", None) is not None or getattr(e, 'header', None) is not None for e in self):
            return df.fillna("").style.hide(axis="index")._repr_html_()
        return df.fillna("")._repr_html_()

    def _border_mask(self, top, right, bottom, left) -> Style:
        return Style(
            first=[top, right, False, left],
            last=[False, right, bottom, left],
            middle=[False, right, False, left],
        )

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.y += elem.height
        return offset

    def _starting_offset(self) -> Loc:
        offset = Loc((0, 0), self._loc.ws)
        if getattr(self, "_header", None) is not None:
            offset.y += 1
        return offset


VFrame.sibling_type = Frame
Frame.sibling_type = VFrame

VFrame.__doc__ = Frame.__doc__

Frame.__doc__ = f"""
    {_Frame._doc_primary_summary}

    {Frame._doc_custom_summary}

    {_Frame._doc_params}

    """

VFrame.__doc__ = f"""
    {_Frame._doc_primary_summary}

    {VFrame._doc_custom_summary}

    {_Frame._doc_params}

    """



from typing import Any

def ensure_value_is_not_number(value: Any, message: str | None = None) -> bool:
    if message is None:
        message = (
            f"Value, '{value}' looks like a number. You've tried to place it in a cell "
            "that must be text to avoid breaking Excel. Please change it to text. "
            "A common example of this is Table headers. "
        )
    try:
        float(value)
        raise TypeError(message)
    except ValueError:
        return True


def ensure_value_is_number(value: Any, message: str | None = None) -> int | float:
    if isinstance(value, (int, float)):
        return value

    if message is None:
        message = (
            f"Value, '{value}' Can't be converted to a number. You've tried to place "
            "it in a cell that must contain a number to avoid breaking Excel. "
            "Please change it to a number."
        )
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            raise ValueError(message)


def require_each_element_to_be_cls_type(container: list) -> None:
    cls_name = type(container).__name__
    elem_type = type(container).elem_type
    elem_type_name = elem_type.__name__
    for i, elem in enumerate(container):
        if not isinstance(elem, elem_type):
            raise TypeError(
                f"Each element inside {cls_name} must be a {elem_type_name}. "
                f"Elem at index {i} is a '{type(elem).__name__}'. Its value is: {elem}"
            )



from typing import Any
from pandas import DataFrame, Series, to_datetime

from excelbird.core.gap import Gap

def get_dimensions(elem: Any) -> int:
    if isinstance(elem, type):
        return getattr(elem, "_dimensions", -1)
    return getattr(type(elem), "_dimensions", -1)


def to_date(column: Series) -> Series:
    return to_datetime(column).dt.date


def datetime_cols_to_dates(df: DataFrame) -> DataFrame:
    """
    Applies `to_date()` to each column in `df` that has a datetime64[ns] type
    """
    df = df.copy()
    for col in df.columns:
        if df[col].dtype == "datetime64[ns]":
            df[col] = to_date(df[col])
    return df


def get_idx(container: list, index: int, default: Any = None) -> Any:
    """
    Safely call list's __getitem__ for an index that might not be valid
    """
    try:
        return container[index]
    except (KeyError, IndexError):
        return default
    

def init_from_same_dimension_type(instance, args: list) -> list:
    """
    Mutates `instance` inplace, AND returns new args.
    """
    first_arg = get_idx(args, 0)
    arg_dimensions = get_dimensions(first_arg)
    instance_dimensions = get_dimensions(instance)
    if (
        len(args) == 1
        and arg_dimensions == instance_dimensions
        and arg_dimensions > 0 and instance_dimensions > 0
    ):
        args = list(first_arg)
        for key, val in first_arg.__dict__.items():
            if key == "_loc":
                continue
            if key == "_id":
                key = "id"
            if key == "_header":
                key = "header"
            setattr(instance, key, val)
    
    return args


def fill_frames(container: list) -> None:
    from excelbird.core.stack import _Stack
    from excelbird.core.frame import _Frame
    from excelbird.core.series import _Series
    from excelbird.core.cell import Cell

    def true_length(series) -> int:
        return len(series) + (0 if series.header is None else 1)

    for elem in container:
        if isinstance(elem, _Stack):
            fill_frames(elem)

        elif isinstance(elem, _Frame):
            if len(set(
                (true_lengths := [
                    true_length(i) for i in elem if isinstance(i, _Series)
                ])
            )) == 1:
                continue

            max_length = max(true_lengths)

            if elem.fill_empty is True:
                fill_value = lambda: Cell("")
            else:
                fill_value = lambda: Gap()

            for series in [e for e in elem if isinstance(e, _Series)]:
                if (true_len := true_length(series)) < max_length:
                    for _ in range(max_length - true_len):
                        series.append(fill_value())


def set_duplicate_objects_to_ref(
    container: list, memory_ids_history: list,
) -> None:
    """
    Duplicated elements need to have .ref() set.
    """
    from excelbird.core.stack import _Stack
    from excelbird.core.frame import _Frame
    from excelbird.core.series import _Series
    from excelbird.core.cell import Cell
    valid_types = (
        _Stack,
        _Frame,
        _Series,
        Cell,
    )
    for i, elem in enumerate(container):
        if isinstance(elem, valid_types):
            if id(elem) in memory_ids_history and hasattr(elem, "ref"):
                container[i] = elem.ref()
            else:
                memory_ids_history.append(id(elem))
                if not isinstance(elem, Cell):
                    set_duplicate_objects_to_ref(elem, memory_ids_history)


def is_notebook() -> bool:
    """
    Thank you very much Gustavo Bezerra on Stackoverflow
    """
    try:
        shell = type(get_ipython()).__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter




# def mark_all_cells_as_written_recursive(container: list) -> None:
#     from excelbird.core.cell import Cell
#     for elem in container:
#         if isinstance(elem, Cell):
#             elem._written = True
#         elif isinstance(elem, list):
#             mark_all_cells_as_written_recursive(elem)






class HasHelp:
    @classmethod
    def help(cls, doc: bool = False):
        doc_str = cls.__doc__
        help_str = cls.__help__
        help_nb_str = cls.__help_notebook__

        if doc_str is None and help_str is None and help_nb_str is None:
            res = f"Sorry, **`{cls.__name__}`** doesn't have any documentation yet."
        elif doc is True or help_str is None:
            res = doc_str
        elif help_str is not None:
            res = help_str
        else:
            res = ""

        from excelbird._utils.util import is_notebook

        if is_notebook():
            from IPython.display import display, Markdown as md

            if help_nb_str is not None:
                display(md(help_nb_str))
            else:
                display(md(res))
        else:
            print(res)

    # Your class's help string
    __help__ = None
    # Custom help string for notebooks
    __help_notebook__ = None



class DotDict(dict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            return super().__getattr__(key)


class Style(DotDict):

    def help(self) -> None:
        print(*self.keys(), sep="\n")



"""
Detailed documentation and code examples coming soon. For now, please use the class
reference page below:
"""
from __future__ import annotations
import re
from typing import TypeVar

from excelbird._layout_references import Globals
from excelbird.styles.styles import default_table_style
from excelbird._base.dotdict import Style
from excelbird._base.math import CanDoMath

from excelbird._utils.pass_attributes import (
    pass_attr_without_override,
    pass_dict_without_override,
)

TExpr = TypeVar("TExpr", bound="Expr")

class Expr(CanDoMath):
    """
    Reference elements in parent container by name or index in a string
    expression. Immediately upon instantiating and Expr, all references
    enclosed in square brackets are parsed out and stored in a dictionary,
    where all values are None. The parent container of an Expr is responsible
    for resolving each of these keys by filling them with their corresponding
    element in the layout, and then evaluating.

    Pass a string containing the code to be executed. It should look like normal python code, but instead
    of variable names, use square brackets and enclose the ``id`` or ``header`` of another layout element
    that may or may not exist yet. Quote marks are *not* necessary inside the brackets. For instance,

    .. code-block::

        Expr("[some_cell_id] / ([some_column_header] + 2)", id="new_col")

    **Shorthand Syntax**: If you don't need to pass any keyword arguments, `Expr` can be more concisely
    written as a single-element :class:`set`, containing the expression string:

    .. code-block::

        {"[some_cell_id] / [other_cell_id]"}

    This is useful when building a :class:`Func`, where you might list many `Expr` together in succession.

    Perhaps you only want to reference a single element, rather than execute an expression. In this
    case, the square brackets are not needed:

    .. code-block::

        {"[some_cell_id]"}
        # Can also be written as
        {"some_cell_id"}

    Again, this shorthand is most useful when building a `Func`

    Parameters
    ----------
    expr_str : str
        The contents of the expression to be executed. It should look like normal python code, but instead
        of variables, use square brackets and enclose an ``id`` or ``header`` of another layout element.
    id : str, optional
        Id of the resulting element once the Expr is evaluated
    header : str, optional
        Header of the resulting element once the Expr is evaluated. This will only have an effect if the
        resulting element can take a header
    cell_style : dict, optional
        To be applied to the resulting element
    header_style : dict, optional
        To be applied to the resulting element
    table_style : dict, optional
        To be applied to the resulting element
    **kwargs : Any
        Any other keyword arguments will be set as attributes on the resulting element

    """

    def __init__(
        self,
        expr_str: str | set,
        id: str | None = None,

        header: str | None = None,

        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        table_style: Style | dict | bool | None = None,
        **kwargs,
    ) -> None:
        if isinstance(expr_str, set):
            expr_str = str(expr_str.pop())
        
        expr_str = expr_str.strip()

        if "[" not in expr_str and "]" not in expr_str:
            expr_str = "[" + expr_str + "]"


        # Match group for the inner contents of a square bracket enclosure
        # that has at least one character and no brackets inside, and is NOT
        # not ONLY digits
        r_elem = r"\[([^\[\]]+?)\]"

        match_start: list = re.findall(r"^" + r_elem, expr_str)

        not_prefixed = r"[^\]\)\[\(]"
        # For all other matches, make sure the enclosure isn't immediately
        # preceded by a bracket or parenthese. Those should
        # be left alone and treated as regular __getitem__ calls in python
        match_others: list = re.findall(not_prefixed + r_elem, expr_str)

        matches = match_start + match_others

        for i, match in enumerate(matches):
            match = match.strip()
            try:
                matches[i] = int(match)
            except Exception:
                try:
                    matches[i] = float(match)
                except Exception:
                    # If match starts AND ends with double quotes, remove them.
                    # Do the same for single quotes
                    matches[i] = re.sub(r'^"(.*?)"$', r"\1", match)
                    matches[i] = re.sub(r"^'(.*?)'$", r"\1", matches[i])

        refs = {match: None for match in matches}

        # Now, our ref names have surrounding quotes removed, so we need
        # to update expr_str to reflect these changes so they match

        r_elem_dub_q = r'\["([^\[\]]+?)"\]'
        r_elem_sing_q = r"\['([^\[\]]+?)'\]"
        cap_not_prefixed = r"(" + not_prefixed + r")"

        expr_str = re.sub(r"^" + r_elem_dub_q, r"[\1]", expr_str)
        expr_str = re.sub(r"^" + r_elem_sing_q, r"[\1]", expr_str)
        expr_str = re.sub(cap_not_prefixed + r_elem_dub_q, r"\1[\2]", expr_str)
        expr_str = re.sub(cap_not_prefixed + r_elem_sing_q, r"\1[\2]", expr_str)

        # Prepare expr_str to be evaluated with eval()
        # - Replace [ref] with self.refs["""ref"""]
        # - Triple quotes must be used since we don't know what kind of quotes, if any,
        #   are already present in the true content of the ref name. It might have both!
        for str_ref in [k for k in refs.keys() if not isinstance(k, (int, float))]:
            expr_str = expr_str.replace(f"[{str_ref}]", f'self.refs["""{str_ref}"""]')

        expr = expr_str

        # Check to see if they're just referencing an object without doing
        # calculations on it. If so, we MIGHT want to return that element's
        # .ref() expression, instead of the element itself. Like if they just
        # list this expression as a cell or a column, we'll want to make a cell
        # reference to that element instead of copying it over. However, if
        # they are referencing an element from inside a function, we want to
        # return the referenced object itself. We'll have to let the parent
        # container determine what to do. If it's a single element reference,
        # we'll set `is_ref` to True right here. The parent container can then
        # set `use_ref` to True for any standalone expressions, not inside functions.
        use_ref = False
        is_ref = False
        if len(refs) == 1:
            if expr.startswith("self") and expr.endswith('"]'):
                is_ref = True

        if cell_style is None: cell_style = dict()
        if header_style is None: header_style = dict()
        if table_style is None or table_style is False: table_style = dict()
        elif table_style is True: table_style = default_table_style

        self.refs = refs
        self._is_ref = is_ref
        self._use_ref = use_ref

        self.expr = expr
        self.expr_str = expr_str
        self.id = id

        self.header = header

        self.cell_style = Style(**cell_style)
        self.header_style = Style(**header_style)
        self.table_style = Style(**table_style)
        self.kwargs = kwargs

    def set(self, **kwargs) -> Expr:
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                self.kwargs[k] = v
        return self

    def _refs_resolved(self) -> bool:
        """
        Whether all references are resolved
        """
        return not any(i is None for i in self.refs.values())

    def _eval(self):
        """
        Call once all references are resolved.
        - Calls python's ``eval()`` on ``self.expr``
        - Applies stored attributes to the result object
        - Returns result object
        """

        if self._refs_resolved() is False:
            raise ValueError("All references must be resolved before calling .eval()")

        res = eval(self.expr)

        if self._use_ref is True:
            res = res.ref()

        # If returning cell, set each style value as attribute on cell
        if getattr(type(res), "_dimensions", None) == 0:
            res._inherit_style_without_override(self.cell_style)
            res.id = self.id

        # If returning _Series, pass down attributes
        if getattr(type(res), "_dimensions", None) >= 1:
            pass_attr_without_override(self, res, "header")
            pass_attr_without_override(self, res, "id")

            pass_dict_without_override(self, res, "cell_style")
            pass_dict_without_override(self, res, "header_style")
            pass_dict_without_override(self, res, "table_style")

            for key, val in self.kwargs.items():
                if hasattr(res, key):
                    if getattr(res, key) is None:
                        setattr(res, key, val)
                elif hasattr(res, "cell_style"):
                    if key not in res.cell_style:
                        res.cell_style[key] = val

        return res

    def _attempt_to_resolve(self, container: list) -> bool:
        """
        Given a container, attempt to resolve all refs

        A valid reference might be:
            An integer index of an item in the container
            A string id in `Globals.ids`

        Returns True if a match was found for each reference

        Mutates inplace: `self`
        """
        from excelbird.core.function import Func
        for key in self.refs.keys():
            try:
                ref = container[container._key_to_idx(key)]
            except (KeyError, IndexError):
                ref = Globals.ids.get(key, None)
                if ref is None:
                    ref = Globals.headers.get(key, None)
                if ref is None:
                    ref = Globals.global_ids.get(key, None)
                if ref is None:
                    ref = Globals.global_headers.get(key, None)

            if ref is not None and not isinstance(ref, (Func, Expr)):
                self.refs[key] = ref
        
        return self._refs_resolved()
    
    @classmethod
    def _resolve_container_recursive(cls, container: list) -> bool:
        """
        For each Expr element, try to resolve its references. If all references
        were resolved, it's replaced with its evaluated form.

        Returns True if all expressions in the container were resolved

        Mutates inplace: `container`
        """
        all_expressions_resolved = True

        for i, elem in enumerate(container):
            if isinstance(elem, cls):

                if elem._attempt_to_resolve(container) is True:
                    container[i] = elem._eval()
                else:
                    all_expressions_resolved = False

            elif isinstance(elem, list):
                if cls._resolve_container_recursive(elem) is False:
                    all_expressions_resolved = False

        return all_expressions_resolved
    
    @classmethod
    def _set_use_ref_for_container_recursive(cls, container: list) -> None:
        """
        Tells all expressions in a container to use references.
        Mutates inplace: `container`
        """
        for elem in container:
            if isinstance(elem, cls):
                if elem._is_ref is True:
                    elem._use_ref = True
            elif isinstance(elem, list):
                cls._set_use_ref_for_container_recursive(elem)
    
    def __repr__(self):
        return f"{type(self).__name__}(...)"





"""
Detailed documentation and code examples coming soon.
"""
from __future__ import annotations
# External
from pandas import Series, DataFrame
from typing import Any
from copy import deepcopy

# Internal main
from excelbird.styles.styles import default_table_style

from excelbird._base.container import ListIndexableById
from excelbird._base.identifier import HasId
from excelbird._base.dotdict import Style
from excelbird._base.loc import Loc
from excelbird._base.styling import (
    HasMargin, 
    HasPadding,
)
from excelbird._utils.util import (
    init_from_same_dimension_type,
)
from excelbird._utils.pass_attributes import (
    pass_attr_to_children,
    pass_dict_to_children,
)
from excelbird._utils.argument_parsing import (
    combine_args_and_children_to_list,
    convert_all_to_type,
    move_remaining_kwargs_to_dict,
)

from excelbird.core.cell import Cell
from excelbird.core.series import (
    _Series,
    Col,
)
from excelbird.core.gap import Gap
from excelbird.core.frame import _Frame, Frame, VFrame
from excelbird.core.expression import Expr


class _Stack(ListIndexableById, HasId, HasMargin, HasPadding):
    _doc_primary_summary = """
    A general container that can hold any element, *including itself*. Offers unique spatial styling
    features, like margin and padding, described below.
    """
    _doc_params = """
    .. note:: Stacks *cannot* be :ref:`used in a python expression <expression_main>`, or included in a :class:`Func`. However, you can still call :meth:`self.ref()` to make an exact reference to its cells.

    Parameters
    ----------
    *args : Union[Stack, VStack, Frame, VFrame, Col, Row, Cell, list, tuple, str, int, float, pd.Series, pd.DataFrame, np.ndarray, Gap, Expr, Func, set]
        Can take any layout element (besides Book or Sheet), or any value that
        can be used to construct a layout element. Stack is the only layout element
        that can store other instances of itself as children
    children : list, optional
        Will be combined with args
    id : str, optional
        Unique identifier to store globally so that this element can be referenced
        elsewhere in the layout without being assigned to a variable
    sep : Gap or bool or int or dict, optional
        A sep in any excelbird layout element inserts a Gap between each of its children.
        If True, a default of Gap(1) is used. If int, Gap(sep) will be used. If a dict,
        ``Gap(1, **sep)`` will be used.
    background_color : str, optional
        Hex code for background_color. Will be applied to fill_color of padding, any Gap
        child who hasn't specified its own fill_color, and to any child Stack/VStack's margins.
        Will also be passed down to any child (Cell excluded) who hasn't specified its own
        background_color.
    schema : Schema, optional
        Applied to each child who takes schema
    cell_style : dict, optional
        Applied to each child who has cell_style
    header_style : dict, optional
        Applied to each child who has header_style
    table_style : dict or bool, optional
        Applied to each child who has table_style
    margin : int or list[int], optional
        Margin, like padding, will apply space around the element. Unlike padding, margin space
        will NOT inherit any of the element's styling. It will, however, be filled with the
        parent container's background_color, if present. Syntax inspired by CSS. An int,
        if passed, will be applied to all 4 sides. If list, length can be 2, 3, or 4 elements.
        Order is [top, right, bottom, left]. If length 2, apply the first element to top and
        bottom margin, and second to right and left.
    margin_top : int, optional
        Top margin, measured in number of cells
    margin_right : int, optional
        Right margin, measured in number of cells
    margin_bottom : int, optional
        Bottom marign, measured in number of cells
    margin_left : int, optional
        Left margin, measured in number of cells
    padding : int or list[int], optional
        Padding, like margin, will apply space around the element. Unlike margin, padding space
        WILL inherit the element's styling, like background_color. Syntax inspired by CSS. An int,
        if passed, will be applied to all 4 sides. If list, length can be 2, 3, or 4 elements.
        Order is [top, right, bottom, left]. If length 2, apply the first element to top and
        bottom margin, and second to right and left.
    padding_top : int, optional
        Top padding, measured in number of cells
    padding_right : int, optional
        Right padding, measured in number of cells
    padding_bottom : int, optional
        Bottom padding, measured in number of cells
    padding_left : int, optional
        Left padding, measured in number of cells
    **kwargs : Any
        Remaining kwargs will be applied to cell_style

    """

    sibling_type = None
    elem_type = None
    _dimensions = -1

    def __init__(
        self,
        *args: Any,
        children: list | None = None,
        id: str | int | None = None,
        sep: Any | None = None,
        background_color: str | None = None,
        margin: int | list[int] | None = None,
        margin_top: int | None = None,
        margin_right: int | None = None,
        margin_bottom: int | None = None,
        margin_left: int | None = None,
        padding: int | list[int] | None = None,
        padding_top: int | None = None,
        padding_right: int | None = None,
        padding_bottom: int | None = None,
        padding_left: int | None = None,
        schema: None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        table_style: Style | dict | bool | None = None,
        **kwargs,
    ) -> None:
        children = combine_args_and_children_to_list(args, children)

        children = [i for i in children if i is not None]

        children = init_from_same_dimension_type(self, children)
        if getattr(self, "_id", None) is not None and id is None:
            id = self.id

        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()
        if table_style is None or table_style is False:
            table_style = dict()
        elif table_style is True:
            table_style = default_table_style

        self._format_args(children)

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        self._loc = None
        self.id = id
        self.background_color = background_color
        # Attrs that must be passed to children
        self.schema = schema
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)
        self.header_style = Style(**header_style)
        self.table_style = Style(**table_style)

        self._init(children)

        self._init_margin(
            margin,
            margin_top,
            margin_right,
            margin_bottom,
            margin_left,
        )
        self._init_padding(
            padding,
            padding_top,
            padding_right,
            padding_bottom,
            padding_left,
        )

        if sep is not None:
            self._insert_separator(sep)


    def ref(self, inherit_style: bool = False, **kwargs):
        """
        Get a new object with cell references to those in the caller.
        This assumes that **both** the calling object
        and the returned object will be placed in the workbook.

        Parameters
        ----------
        inherit_style : bool, default False
            Copy the caller's style to the returned object.

        Returns
        -------
        :class:`Stack <excelbird.Stack>` or :class:`VStack <excelbird.VStack>`
            Self type

        Notes
        -----

        .. note::

            Children's ``header`` attributes are stylistic attributes, and therefore will **not** be
            passed to the returned object's children unless ``inherit_style=True``. And, if style
            is inherited, headers will be copied over to the children, instead of cell references to them.

        """
        new_elements = [
            i.ref(inherit_style=inherit_style, **kwargs)
            if not isinstance(i, Gap)
            else deepcopy(i)
            for i in self
        ]
        new_dict = kwargs
        if inherit_style is True:
            self_dict = deepcopy(self.__dict__)
            for key, val in self_dict.items():
                if key == "_header":
                    key = "header"
                if key not in new_dict and key not in ["_id", "_loc"]:
                    new_dict[key] = val
        return type(self)(*new_elements, **new_dict)

    def transpose(self, **kwargs):
        """
        Convert to sibling type. Places current children into the returned object,
        without copying or making cell references to them.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to apply as attributes to the new object.

        Returns
        -------
        :class:`Stack <excelbird.Stack>` or :class:`VStack <excelbird.VStack>`
            The opposite to self's type. Try ``type(my_obj).sibling_type``

        Notes
        -----
        **Assumes that the caller won't be placed in the layout**. Do not
        place both the calling object and returned object in the layout, since
        they both contain the same children.

        .. code-block::

            # 'current' must not be placed in the workbook.
            new = current.transpose()

        To include the caller and make cell references to it, get a reference
        first:

        .. code-block::

            new = current.ref().transpose()

        """
        elements = list(self)
        new = type(self).sibling_type(*elements)
        for key, val in self.__dict__.items():
            if key == "_id":
                key = "id"
            setattr(new, key, val)
        for key, val in kwargs.items():
            if hasattr(new, key):
                setattr(new, key, val)
            elif hasattr(new, 'cell_style'):
                new.cell_style[key] = val
        return new

    def _format_args(self, args: list) -> None:
        convert_all_to_type(args, (str, int, float), Cell, strict=True)
        convert_all_to_type(args, Series, Col)
        convert_all_to_type(args, DataFrame, Frame)
        convert_all_to_type(args, set, Expr)

    @property
    def _elem_widths(self) -> list:
        return [i.width for i in self if hasattr(i, "width")]

    @property
    def _elem_heights(self) -> list:
        return [i.height for i in self if hasattr(i, "height")]

    def _resolve_background_color(self) -> None:

        for elem in self:
            if hasattr(elem, "_resolve_background_color"):
                if (
                    self.background_color not in [None, False]
                    and elem.background_color is None
                ):
                    elem.background_color = self.background_color
                elem._resolve_background_color()

        if self.background_color not in [None, False]:
            for elem in self:
                if isinstance(elem, Gap):
                    if "fill_color" not in elem.kwargs and elem.is_margin is False:
                        elem.fill = True
                        elem.kwargs["fill_color"] = self.background_color

                # Child's margins should be filled with self's background color
                elif hasattr(elem, "margin"):
                    if elem.margin != HasMargin.empty:
                        for item in elem:
                            if isinstance(item, Gap):
                                if (
                                    "fill_color" not in item.kwargs
                                    and item.is_margin is True
                                ):
                                    item.fill = True
                                    item.kwargs["fill_color"] = self.background_color
                            elif hasattr(item, "margin"):
                                for x in item:
                                    if isinstance(x, Gap):
                                        if (
                                            "fill_color" not in x.kwargs
                                            and x.is_margin is True
                                        ):
                                            x.fill = True
                                            x.kwargs[
                                                "fill_color"
                                            ] = self.background_color

    def _resolve_padding(self) -> None:
        for elem in self:
            if hasattr(elem, "padding"):
                elem._resolve_padding()

        def get_gap(amount, elem) -> Gap:
            if getattr(elem, "background_color", None) not in [None, False]:
                return Gap(amount, fill_color=elem.background_color)
            return Gap(amount, fill_color=None)

        for i, elem in enumerate(self):
            if hasattr(elem, "padding"):
                if elem.padding != HasPadding.empty:
                    top, right, bottom, left = elem.padding
                    elem_type = type(elem)
                    new_elements = []
                    for i, item in reversed(list(enumerate(elem))):
                        new_elements.insert(0, elem.pop(i))

                    if issubclass(elem_type, Stack):
                        if left is not None:
                            elem.append(get_gap(left, elem))

                        elem.append(
                            elem_type.sibling_type(
                                get_gap(top, elem) if top is not None else None,
                                elem_type(*new_elements),
                                get_gap(bottom, elem) if bottom is not None else None,
                            )
                        )
                        if right is not None:
                            elem.append(get_gap(right, elem))

                    elif issubclass(elem_type, VStack):
                        if top is not None:
                            elem.append(get_gap(top, elem))

                        elem.append(
                            elem_type.sibling_type(
                                get_gap(left, elem) if left is not None else None,
                                elem_type(*new_elements),
                                get_gap(right, elem) if right is not None else None,
                            ),
                        )
                        if bottom is not None:
                            elem.append(get_gap(bottom, elem))

    def _resolve_margin(self) -> None:
        for elem in self:
            if hasattr(elem, "margin"):
                elem._resolve_margin()

        for i, elem in enumerate(self):
            if hasattr(elem, "margin"):
                if elem.margin != HasMargin.empty:
                    top, right, bottom, left = elem.margin
                    elem_type = type(elem)
                    new_elements = []
                    for i, item in reversed(list(enumerate(elem))):
                        new_elements.insert(0, elem.pop(i))

                    if issubclass(elem_type, Stack):
                        if left is not None:
                            elem.append(Gap(left, is_margin=True))

                        elem.append(
                            elem_type.sibling_type(
                                Gap(top, is_margin=True) if top is not None else None,
                                elem_type(*new_elements),
                                Gap(bottom, is_margin=True)
                                if bottom is not None
                                else None,
                            ),
                        )
                        if right is not None:
                            elem.append(Gap(right, is_margin=True))

                    elif issubclass(elem_type, VStack):
                        if top is not None:
                            elem.append(Gap(top, is_margin=True))

                        elem.append(
                            elem_type.sibling_type(
                                Gap(left, is_margin=True) if left is not None else None,
                                elem_type(*new_elements),
                                Gap(right, is_margin=True)
                                if right is not None
                                else None,
                            ),
                        )
                        if bottom is not None:
                            elem.append(Gap(bottom, is_margin=True))

    def _resolve_gaps(self) -> None:
        Gap._convert_all_to_frames(self, type(self).elem_type, self._gap_size)
        for elem in self:
            if hasattr(elem, "_resolve_gaps"):
                elem._resolve_gaps()

    def _set_loc(self, loc: Loc) -> None:
        self._loc = loc

        offset = self._starting_offset()
        for elem in self:
            elem._set_loc(
                Loc((self._loc.y + offset.y, self._loc.x + offset.x), self._loc.ws)
            )
            offset = self._inc_offset(offset, elem)

    def __getitem__(self, key):
        if not isinstance(key, list):
            # return super().__getitem__(key)
            return ListIndexableById.__getitem__(self, key)

        new_elements = [self[self._key_to_idx(k)] for k in key]
        new_dict = {k: v for k, v in self.__dict__.items() if k not in ["_id", "_loc"]}
        if "_header" in new_dict:
            new_dict["header"] = new_dict.pop("_header")

        return type(self)(*new_elements, **new_dict)

    def _validate_child_types(self) -> None:
        valid_types = (
            _Stack,
            _Frame,
            _Series,
            Cell,
            Gap,
        )
        for elem in self:
            if not isinstance(elem, valid_types):
                raise TypeError(
                    f"At write time, a {type(self).__name__} can only hold "
                    "the following types:\n{valid_types}"
                )
            if hasattr(elem, "_validate_child_types"):
                elem._validate_child_types()

    def _write(self) -> None:
        pass_attr_to_children(self, "schema")
        pass_dict_to_children(self, "cell_style")
        pass_dict_to_children(self, "header_style")
        pass_dict_to_children(self, "table_style")

        if len(self.cell_style) > 0:
            for elem in self:
                if isinstance(elem, Cell):
                    elem._inherit_style_without_override(self.cell_style)

        for elem in self:
            elem._write()


class Stack(_Stack):
    _doc_custom_summary = """
    * Direction: **horizontal**
    * Child Type: :class:`Stack`, :class:`VStack`, :class:`Frame`, :class:`VFrame`, :class:`Col`, :class:`Row`, :class:`Cell`
    """
    sibling_type = None  # these are set after class declaration
    elem_type = Frame

    def transpose(self, **kwargs) -> VStack:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> Stack:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self) -> int:
        return sum(self._elem_widths + [0])

    @property
    def height(self) -> int:
        heights = [i.height for i in self if hasattr(i, 'height') and not isinstance(i, Gap)]
        return max(heights + [0])

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.x += elem.width
        return offset

    def _starting_offset(self) -> Loc:
        return Loc((0, 0), self._loc.ws)

    @property
    def _gap_size(self) -> int:
        return self.height


class VStack(_Stack):
    _doc_custom_summary = """
    * Direction: **vertical**
    * Child Type: :class:`Stack`, :class:`VStack`, :class:`Frame`, :class:`VFrame`, :class:`Col`, :class:`Row`, :class:`Cell`
    """
    sibling_type = None  # these are set after class declaration
    elem_type = VFrame

    def transpose(self, **kwargs) -> Stack:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> VStack:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self) -> int:
        widths = [i.width for i in self if hasattr(i, 'width') and not isinstance(i, Gap)]
        return max(widths + [0])

    @property
    def height(self) -> int:
        return sum(self._elem_heights + [0])

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.y += elem.height
        return offset

    def _starting_offset(self) -> Loc:
        return Loc((0, 0), self._loc.ws)

    @property
    def _gap_size(self) -> int:
        return self.width


Stack.sibling_type = VStack
VStack.sibling_type = Stack

Stack.__doc__ = f"""
    {_Stack._doc_primary_summary}

    {Stack._doc_custom_summary}

    {_Stack._doc_params}

    """

VStack.__doc__ = f"""
    {_Stack._doc_primary_summary}

    {VStack._doc_custom_summary}

    {_Stack._doc_params}

    """



from excelbird._base.dotdict import Style

conditional = Style(
    light_green="C6EFCE",
    dark_green="006100",
    light_yellow="FFEB9C",
    dark_yellow="9C6500",
    light_red="FFC7CE",
    dark_red="9C0006",
)
mono = Style(
    white="FFFFFF",
    gray9="1A1A1A",  # 10%
    gray8="333333",  # 20%
    gray7="4D4D4D",  # 30%
    gray6="666666",  # 40%
    gray5="808080",  # 50%
    gray4="999999",  # 60%
    gray3="B3B3B3",  # 70%
    gray2="CCCCCC",  # 80%
    gray1="E6E6E6",  # 90%
    gray0="F3F3F3",  # 95%
    black="000000",
)
standard = Style(
    dark_red="C00000",
    red="FF0000",
    orange="FFC000",
    yellow="FFFF00",
    light_green="92D050",
    green="00B050",
    light_blue="00B0F0",
    blue="0070C0",
    dark_blue="002060",
    purple="7030A0",
)
theme = Style(
    light="FFFFFF",
    light1="F2F2F2",
    light2="D9D9D9",
    light3="BFBFBF",
    light4="A6A6A6",
    light5="808080",
    dark="000000",
    dark1="808080",
    dark2="595959",
    dark3="404040",
    dark4="262626",
    dark5="0D0D0D",
    tan="EEECE1",
    tan1="DDD9C4",
    tan2="C4BD97",
    tan3="948A54",
    tan4="494529",
    tan5="1D1B10",
    dark_blue="1F497D",
    dark_blue1="C5D9F1",
    dark_blue2="8DB4E2",
    dark_blue3="538DD5",
    dark_blue4="16365C",
    dark_blue5="0F243E",
    light_blue="4F81BD",
    light_blue1="DCE6F1",
    light_blue2="B8CCE4",
    light_blue3="95B3D7",
    light_blue4="366092",
    light_blue5="244062",
    red="C0504D",
    red1="F2DCDB",
    red2="E6B8B7",
    red3="DA9694",
    red4="963634",
    red5="632523",
    green="9BBB59",
    green1="EBF1DE",
    green2="D8E4BC",
    green3="C4D79B",
    green4="76933C",
    green5="4F6228",
    purple="8064A2",
    purple1="E4DFEC",
    purple2="CCC0DA",
    purple3="B1A0C7",
    purple4="60497A",
    purple5="403151",
    aqua="4BACC6",
    aqua1="DAEEF3",
    aqua2="B7DEE8",
    aqua3="92CDDC",
    aqua4="31869B",
    aqua5="215967",
    orange="F79646",
    orange1="FDE9D9",
    orange2="FCD5B4",
    orange3="FABF8F",
    orange4="E26B0A",
    orange5="974706",
)

theme["groups"] = Style(
    light=[
        theme.light,
        theme.light1,
        theme.light2,
        theme.light3,
        theme.light4,
        theme.light5,
    ],
    dark=[
        theme.dark,
        theme.dark1,
        theme.dark2,
        theme.dark3,
        theme.dark4,
        theme.dark5,
    ],
    tan=[
        theme.tan,
        theme.tan1,
        theme.tan2,
        theme.tan3,
        theme.tan4,
        theme.tan5,
    ],
    dark_blue=[
        theme.dark_blue,
        theme.dark_blue1,
        theme.dark_blue2,
        theme.dark_blue3,
        theme.dark_blue4,
        theme.dark_blue5,
    ],
    light_blue=[
        theme.light_blue,
        theme.light_blue1,
        theme.light_blue2,
        theme.light_blue3,
        theme.light_blue4,
        theme.light_blue5,
    ],
    red=[
        theme.red,
        theme.red1,
        theme.red2,
        theme.red3,
        theme.red4,
        theme.red5,
    ],
    green=[
        theme.green,
        theme.green1,
        theme.green2,
        theme.green3,
        theme.green4,
        theme.green5,
    ],
    purple=[
        theme.purple,
        theme.purple1,
        theme.purple2,
        theme.purple3,
        theme.purple4,
        theme.purple5,
    ],
    aqua=[
        theme.aqua,
        theme.aqua1,
        theme.aqua2,
        theme.aqua3,
        theme.aqua4,
        theme.aqua5,
    ],
    orange=[
        theme.orange,
        theme.orange1,
        theme.orange2,
        theme.orange3,
        theme.orange4,
        theme.orange5,
    ],
)



from typing import Any

def pass_attr_to_children(container: list, attr_name: str) -> None:
    """
    For each child element whose attribute, `attr_name` is not None,
    set that attribute to the value of the parent's `attr_name`

    Mutates inplace: `container`
    """
    container_attr = getattr(container, attr_name)
    if container_attr is None:
        return
    for elem in container:
        if hasattr(elem, attr_name):
            if getattr(elem, attr_name) is None:
                setattr(elem, attr_name, container_attr)


def pass_dict_to_children(container: list, dict_name: str) -> None:
    """
    Copy each element in a parent container's `dict_name` to each
    of its children's `dict_name` if that child hasn't already set it.
    ---
    Container MUST have an attribute called `dict_name`, and that
    attribute MUST be an instance of dictionary.

    Mutates inplace: `container`
    """
    container_dict = getattr(container, dict_name)
    if container_dict is None:
        return
    if len(container_dict) == 0:
        return
    for elem in container:
        if hasattr(elem, dict_name):
            for attr, value in container_dict.items():
                if attr not in getattr(elem, dict_name):
                    getattr(elem, dict_name)[attr] = value


def pass_attr_without_override(elem1: Any, elem2: Any, attr_name: str) -> None:
    """
    If elem1 and elem2 have attribute, `attr_name` and that attribute is not None
    for elem1 and IS None for elem2, set elem1's value for the attribute to elem2

    Mutates inplace:
        `elem1`
        `elem2`
    """
    assert hasattr(elem1, attr_name), (
        f"Element 1 must have attribute, '{attr_name}'"
    )
    if not hasattr(elem2, attr_name):
        return

    elem1_attr = getattr(elem1, attr_name)
    elem2_attr = getattr(elem2, attr_name)

    if elem1_attr is None:
        return
    if elem2_attr is not None:
        return
    
    setattr(elem2, attr_name, elem1_attr)


def pass_dict_without_override(elem1: Any, elem2: Any, dict_name: str) -> None:
    """
    If elem1 and elem2 have attribute, `dict_name` and that attribute is a dict,
    Safely pass each each element from `elem1`'s `dict_name` to `elem2`'s `dict_name`

    Mutates inplace:
        `elem1`
        `elem2`
    """
    assert hasattr(elem1, dict_name), (
        f"Element 1 must have attribute, {dict_name}"
    )
    if not hasattr(elem2, dict_name):
        return

    elem1_dict = getattr(elem1, dict_name)
    elem2_dict = getattr(elem2, dict_name)

    assert (
        elem1_dict is not None and elem2_dict is not None
    ), f"Both element's {dict_name} can't be None"

    for key, val in elem1_dict.items():
        if key not in elem2_dict:
            elem2_dict[key] = val




# These are the names of all excel functions, as they should appear
# in Excel. We use these for validating
FORMULAE = {
	"ABS",
	"ACCRINT",
	"ACCRINTM",
	"ACOS",
	"ACOSH",
	"ACOT",
	"ACOTH",
	"AGGREGATE",
	"ADDRESS",
	"AMORDEGRC",
	"AMORLINC",
	"AND",
	"ARABIC",
	"AREAS",
	"ARRAYTOTEXT",
	"ASC",
	"ASIN",
	"ASINH",
	"ATAN",
	"ATAN2",
	"ATANH",
	"AVEDEV",
	"AVERAGE",
	"AVERAGEA",
	"AVERAGEIF",
	"AVERAGEIFS",
	"BAHTTEXT",
	"BASE",
	"BESSELI",
	"BESSELJ",
	"BESSELK",
	"BESSELY",
	"BETADIST",
	"BETA.DIST",
	"BETAINV",
	"BETA.INV",
	"BIN2DEC",
	"BIN2HEX",
	"BIN2OCT",
	"BINOMDIST",
	"BINOM.DIST",
	"BINOM.DIST.RANGE",
	"BINOM.INV",
	"BITAND",
	"BITLSHIFT",
	"BITOR",
	"BITRSHIFT",
	"BITXOR",
	"BYCOL",
	"BYROW",
	"CALL",
	"CEILING",
	"CEILING.MATH",
	"CEILING.PRECISE",
	"CELL",
	"CHAR",
	"CHIDIST",
	"CHIINV",
	"CHITEST",
	"CHISQ.DIST",
	"CHISQ.DIST.RT",
	"CHISQ.INV",
	"CHISQ.INV.RT",
	"CHISQ.TEST",
	"CHOOSE",
	"CHOOSECOLS",
	"CHOOSEROWS",
	"CLEAN",
	"CODE",
	"COLUMN",
	"COLUMNS",
	"COMBIN",
	"COMBINA",
	"CONCAT",
	"CONCATENATE",
	"CONFIDENCE",
	"CONFIDENCE.NORM",
	"CONFIDENCE.T",
	"CONVERT",
	"CORREL",
	"COS",
	"COSH",
	"COT",
	"COTH",
	"COUNT",
	"COUNTA",
	"COUNTBLANK",
	"COUNTIF",
	"COUNTIFS",
	"COUPDAYBS",
	"COUPDAYS",
	"COUPDAYSNC",
	"COUPNCD",
	"COUPNUM",
	"COUPPCD",
	"COVAR",
	"COVARIANCE.P",
	"COVARIANCE.S",
	"CRITBINOM",
	"CSC",
	"CSCH",
	"CUBEKPIMEMBER",
	"CUBEMEMBER",
	"CUBEMEMBERPROPERTY",
	"CUBERANKEDMEMBER",
	"CUBESET",
	"CUBESETCOUNT",
	"CUBEVALUE",
	"CUMIPMT",
	"CUMPRINC",
	"DATE",
	"DATEDIF",
	"DATEVALUE",
	"DAVERAGE",
	"DAY",
	"DAYS",
	"DAYS360",
	"DB",
	"DBCS",
	"DCOUNT",
	"DCOUNTA",
	"DDB",
	"DEC2BIN",
	"DEC2HEX",
	"DEC2OCT",
	"DECIMAL",
	"DEGREES",
	"DELTA",
	"DEVSQ",
	"DGET",
	"DISC",
	"DMAX",
	"DMIN",
	"DOLLAR",
	"DOLLARDE",
	"DOLLARFR",
	"DPRODUCT",
	"DROP",
	"DSTDEV",
	"DSTDEVP",
	"DSUM",
	"DURATION",
	"DVAR",
	"DVARP",
	"EDATE",
	"EFFECT",
	"ENCODEURL",
	"EOMONTH",
	"ERF",
	"ERF.PRECISE",
	"ERFC",
	"ERFC.PRECISE",
	"ERROR.TYPE",
	"EUROCONVERT",
	"EVEN",
	"EXACT",
	"EXP",
	"EXPAND",
	"EXPON.DIST",
	"EXPONDIST",
	"FACT",
	"FACTDOUBLE",
	"FALSE",
	"F.DIST",
	"FDIST",
	"F.DIST.RT",
	"FILTER",
	"FILTERXML",
	"FIND",
	"FINDB",
	"F.INV",
	"F.INV.RT",
	"FINV",
	"FISHER",
	"FISHERINV",
	"FIXED",
	"FLOOR",
	"FLOOR.MATH",
	"FLOOR.PRECISE",
	"FORECAST",
	"FORMULATEXT",
	"FREQUENCY",
	"F.TEST",
	"FTEST",
	"FV",
	"FVSCHEDULE",
	"GAMMA",
	"GAMMA.DIST",
	"GAMMADIST",
	"GAMMA.INV",
	"GAMMAINV",
	"GAMMALN",
	"GAMMALN.PRECISE",
	"GAUSS",
	"GCD",
	"GEOMEAN",
	"GESTEP",
	"GETPIVOTDATA",
	"GROWTH",
	"HARMEAN",
	"HEX2BIN",
	"HEX2DEC",
	"HEX2OCT",
	"HLOOKUP",
	"HOUR",
	"HSTACK",
	"HYPERLINK",
	"HYPGEOM.DIST",
	"HYPGEOMDIST",
	"IF",
	"IFERROR",
	"IFNA",
	"IFS",
	"IMABS",
	"IMAGINARY",
	"IMARGUMENT",
	"IMCONJUGATE",
	"IMCOS",
	"IMCOSH",
	"IMCOT",
	"IMCSC",
	"IMCSCH",
	"IMDIV",
	"IMEXP",
	"IMLN",
	"IMLOG10",
	"IMLOG2",
	"IMPOWER",
	"IMPRODUCT",
	"IMREAL",
	"IMSEC",
	"IMSECH",
	"IMSIN",
	"IMSINH",
	"IMSQRT",
	"IMSUB",
	"IMSUM",
	"IMTAN",
	"INDEX",
	"INDIRECT",
	"INFO",
	"INT",
	"INTERCEPT",
	"INTRATE",
	"IPMT",
	"IRR",
	"ISBLANK",
	"ISERR",
	"ISERROR",
	"ISEVEN",
	"ISFORMULA",
	"ISLOGICAL",
	"ISNA",
	"ISNONTEXT",
	"ISNUMBER",
	"ISODD",
	"ISOMITTED",
	"ISREF",
	"ISTEXT",
	"ISO.CEILING",
	"ISOWEEKNUM",
	"ISPMT",
	"JIS",
	"KURT",
	"LAMBDA",
	"LARGE",
	"LCM",
	"LEFT",
	"LEFTB",
	"LEN",
	"LENB",
	"LET",
	"LINEST",
	"LN",
	"LOG",
	"LOG10",
	"LOGEST",
	"LOGINV",
	"LOGNORM.DIST",
	"LOGNORMDIST",
	"LOGNORM.INV",
	"LOOKUP",
	"LOWER",
	"MAKEARRAY",
	"MAP",
	"MATCH",
	"MAX",
	"MAXA",
	"MAXIFS",
	"MDETERM",
	"MDURATION",
	"MEDIAN",
	"MID",
	"MIDB",
	"MIN",
	"MINIFS",
	"MINA",
	"MINUTE",
	"MINVERSE",
	"MIRR",
	"MMULT",
	"MOD",
	"MODE",
	"MODE.MULT",
	"MODE.SNGL",
	"MONTH",
	"MROUND",
	"MULTINOMIAL",
	"MUNIT",
	"N",
	"NA",
	"NEGBINOM.DIST",
	"NEGBINOMDIST",
	"NETWORKDAYS",
	"NETWORKDAYS.INTL",
	"NOMINAL",
	"NORM.DIST",
	"NORMDIST",
	"NORMINV",
	"NORM.INV",
	"NORM.S.DIST",
	"NORMSDIST",
	"NORM.S.INV",
	"NORMSINV",
	"NOT",
	"NOW",
	"NPER",
	"NPV",
	"NUMBERVALUE",
	"OCT2BIN",
	"OCT2DEC",
	"OCT2HEX",
	"ODD",
	"ODDFPRICE",
	"ODDFYIELD",
	"ODDLPRICE",
	"ODDLYIELD",
	"OFFSET",
	"OR",
	"PDURATION",
	"PEARSON",
	"PERCENTILE.EXC",
	"PERCENTILE.INC",
	"PERCENTILE",
	"PERCENTRANK.EXC",
	"PERCENTRANK.INC",
	"PERCENTRANK",
	"PERMUT",
	"PERMUTATIONA",
	"PHI",
	"PHONETIC",
	"PI",
	"PMT",
	"POISSON.DIST",
	"POISSON",
	"POWER",
	"PPMT",
	"PRICE",
	"PRICEDISC",
	"PRICEMAT",
	"PROB",
	"PRODUCT",
	"PROPER",
	"PV",
	"QUARTILE",
	"QUARTILE.EXC",
	"QUARTILE.INC",
	"QUOTIENT",
	"RADIANS",
	"RAND",
	"RANDARRAY",
	"RANDBETWEEN",
	"RANK.AVG",
	"RANK.EQ",
	"RANK",
	"RATE",
	"RECEIVED",
	"REDUCE",
	"REGISTER.ID",
	"REPLACE",
	"REPLACEB",
	"REPT",
	"RIGHT",
	"RIGHTB",
	"ROMAN",
	"ROUND",
	"ROUNDDOWN",
	"ROUNDUP",
	"ROW",
	"ROWS",
	"RRI",
	"RSQ",
	"RTD",
	"SCAN",
	"SEARCH",
	"SEARCHB",
	"SEC",
	"SECH",
	"SECOND",
	"SEQUENCE",
	"SERIESSUM",
	"SHEET",
	"SHEETS",
	"SIGN",
	"SIN",
	"SINH",
	"SKEW",
	"SKEW.P",
	"SLN",
	"SLOPE",
	"SMALL",
	"SORT",
	"SORTBY",
	"SQRT",
	"SQRTPI",
	"STANDARDIZE",
	"STOCKHISTORY",
	"STDEV",
	"STDEV.P",
	"STDEV.S",
	"STDEVA",
	"STDEVP",
	"STDEVPA",
	"STEYX",
	"SUBSTITUTE",
	"SUBTOTAL",
	"SUM",
	"SUMIF",
	"SUMIFS",
	"SUMPRODUCT",
	"SUMSQ",
	"SUMX2MY2",
	"SUMX2PY2",
	"SUMXMY2",
	"SWITCH",
	"SYD",
	"T",
	"TAN",
	"TANH",
	"TAKE",
	"TBILLEQ",
	"TBILLPRICE",
	"TBILLYIELD",
	"T.DIST",
	"T.DIST.2T",
	"T.DIST.RT",
	"TDIST",
	"TEXT",
	"TEXTAFTER",
	"TEXTBEFORE",
	"TEXTJOIN",
	"TEXTSPLIT",
	"TIME",
	"TIMEVALUE",
	"T.INV",
	"T.INV.2T",
	"TINV",
	"TOCOL",
	"TOROW",
	"TODAY",
	"TRANSPOSE",
	"TREND",
	"TRIM",
	"TRIMMEAN",
	"TRUE",
	"TRUNC",
	"T.TEST",
	"TTEST",
	"TYPE",
	"UNICHAR",
	"UNICODE",
	"UNIQUE",
	"UPPER",
	"VALUE",
	"VALUETOTEXT",
	"VAR",
	"VAR.P",
	"VAR.S",
	"VARA",
	"VARP",
	"VARPA",
	"VDB",
	"VLOOKUP",
	"VSTACK",
	"WEBSERVICE",
	"WEEKDAY",
	"WEEKNUM",
	"WEIBULL",
	"WEIBULL.DIST",
	"WORKDAY",
	"WORKDAY.INTL",
	"WRAPCOLS",
	"WRAPROWS",
	"XIRR",
	"XLOOKUP",
	"XMATCH",
	"XNPV",
	"XOR",
	"YEAR",
	"YEARFRAC",
	"YIELD",
	"YIELDDISC",
	"YIELDMAT",
	"Z.TEST",
	"ZTEST",
}



class Globals:
    ids = dict()
    headers = dict()
    global_ids = dict()
    global_headers = dict()
    force_valid_references = True
    expression_sign_spacing = 2

    @classmethod
    def clear_references(cls, sheet_title: str | None = None) -> None:
        if sheet_title is None:
            cls.ids = dict()
            cls.headers = dict()
            return

        ids_to_pop = []
        headers_to_pop = []

        for id, elem in cls.ids.items():
            if hasattr(elem, "loc"):
                if elem._loc is not None:
                    if elem._loc.ws.title == sheet_title:
                        ids_to_pop.append(id)

        for header, elem in cls.headers.items():
            if hasattr(elem, "_loc"):
                if elem._loc is not None:
                    if elem._loc.ws.title == sheet_title:
                        headers_to_pop.append(header)

        for header in headers_to_pop:
            cls.headers.pop(header)
        for id in ids_to_pop:
            cls.ids.pop(id)

    @classmethod
    def clear_global_references(cls) -> None:
        cls.global_ids = dict()
        cls.global_headers = dict()
